"""app providers the server and its services, as well as script entry."""

import logistro

from tetsuya._globals import cli

from . import server as server  # need this to be executed
from . import services as services  # need this to be executed

_logger = logistro.getLogger(__name__)


def main():  # script entry point
    """Start the cli service."""
    _, remaining = logistro.parser.parse_known_args()
    cli(args=remaining)



import asyncio

import logistro
import pytest

# allows to create a browser pool for tests
pytestmark = pytest.mark.asyncio(loop_scope="function")

_logger = logistro.getLogger(__name__)


async def test_placeholder():
    _logger.info("testing placeholder")
    await asyncio.sleep(0)
    assert True



"""Contains helper functions for the client. Only utils dep."""

import atexit
from pathlib import Path

import httpx

from .utils import uds_path


def get_client(
    path: Path | None = None,
    *,
    defer_close=True,
    timeout=0.05,
) -> httpx.Client:
    """Get a client you can use for executing commands."""
    transport = httpx.HTTPTransport(uds=str(path or uds_path()))
    client = httpx.Client(
        timeout=httpx.Timeout(timeout),
        transport=transport,
        base_url="http://tetsuya",
    )
    if defer_close:
        atexit.register(client.close)
    return client



"""These globals, needed everywhere, cannot depend on any tetsuya thing."""

from __future__ import annotations

from typing import TYPE_CHECKING

import logistro
import orjson
import typer
from fastapi import FastAPI
from fastapi.responses import Response

if TYPE_CHECKING:
    from .app.services._base import Bannin

_logger = logistro.getLogger(__name__)


class ORJSONUtcResponse(Response):
    media_type = "application/json"

    def render(self, content) -> bytes:
        return orjson.dumps(
            content,
            option=(
                orjson.OPT_NAIVE_UTC  # treat naive datetimes as UTC
                | orjson.OPT_UTC_Z  # use "Z" instead of +00:00
                | orjson.OPT_SERIALIZE_DATACLASS
                | orjson.OPT_SERIALIZE_NUMPY
            ),
        )


cli = typer.Typer(help="tetsuya CLI")
# Our server daemon
app = FastAPI(title="Tetsuya", default_response_class=ORJSONUtcResponse)

# A list of possible services
service_types: list[type[Bannin]] = []

# A list of running services, only activated by start
active_services: dict[str, Bannin] = {}



"""Bottom dependency."""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Generic, TypeVar

import logistro

from tetsuya.timer import reschedule

from .utils.config import config_data

if TYPE_CHECKING:
    from typing import Any


_logger = logistro.getLogger(__name__)

@dataclass(slots=True)
class Settei(ABC):
    cachelife: int = 0 # number of seconds
    autorefresh: bool = False

    @classmethod
    def default_config(cls) -> dict[str, Any]:
        return asdict(cls())

TSettei = TypeVar("TSettei", bound=Settei)

class Tsuho(ABC):
    """The object a service stores or returns."""

    @abstractmethod
    def long(self) -> str: ...

    @abstractmethod
    def short(self) -> str: ...

    created_at: datetime | None = None

    def _is_stamped(self) -> bool:
        return bool(hasattr(self, "created_at") and self.created_at) is not None

    def since_when(self) -> timedelta | None:
        if not self._is_stamped() or self.created_at is None: # redundant but typechecker dumb
            return None
        return datetime.now(tz=UTC) - self.created_at

    def tstamp(self) -> None:
        self.create_at = datetime.now(tz=UTC)

    def is_live(self, config: Settei | None) -> bool:
        if not self._is_stamped() or self.created_at is None:
            return False
        if not config or not hasattr(config, "cachelife"):
            return False
        return self.created_at + timedelta(seconds=config.cachelife) > datetime.now(tz=UTC)

class Bannin(ABC, Generic[TSettei]):
    """The abstract idea of a service."""

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """Get name of service."""

    @classmethod
    @abstractmethod
    def get_config_type(cls) -> type[TSettei]:
        """Get a config dataclass."""

    cache: Tsuho | None

    @abstractmethod
    def _execute(self) -> Tsuho: ...

    def get_report(self) -> Tsuho | None:
        """Get the actual latest result object."""
        if not hasattr(self, "cache"):
            self.cache = None
        return self.cache

    async def run(self, *, force=False):
        """Run the service in a cache-aware manner."""
        _logger.info(f"Running {self.get_name()}")
        cache = self.get_report()
        if (
                not force
                and hasattr(cache, "is_live") and cache
                and cache.is_live(config=self.get_config())
                ):
            _logger.info("Not rerunning- cache is live.")
            return
        cache = await asyncio.to_thread(self._execute)
        reschedule(self)
        if hasattr(cache, "tstamp") and cache:
            cache.tstamp()
        self.cache = cache

        _logger.debug2(f"New cache: {self.cache}")

    @classmethod
    def get_config(cls) -> TSettei:
        return cls.get_config_type()(**config_data.get(cls.get_name(), {}))



from __future__ import annotations

import asyncio
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, TypedDict

import logistro

if TYPE_CHECKING:
    from .app.services._base import Bannin

class TaskData(TypedDict):
    task: asyncio.Task
    cachelife: int
    tstamp: datetime
    service: Bannin

timer_tasks: dict[str, TaskData] = {}

_logger = logistro.getLogger(__name__)

def reconfig(service: Bannin):
    cfg = service.get_config()
    if not cfg.autorefresh:
        deschedule(service)
    else:
        if not (_td := timer_tasks.get(service.get_name())):
            reschedule(service, for_when=0)
        else:
            if _td["cachelife"] != cfg.cachelife:
                duedate = _td["tstamp"] + timedelta(seconds=cfg.cachelife)
                if duedate <= datetime.now(tz=UTC):
                    reschedule(_td["service"], for_when=0)
                else:
                    for_when = (duedate - datetime.now(tz=UTC)).total_seconds()
                    reschedule(_td["service"], for_when=int(for_when))

def _clear_task(t: asyncio.Task):
    if t.cancelled():
        return
    else:
        if (e := t.exception()):
            _logger.error("Error while timer clears a task.", exc_info=e)

def _post_task(service: Bannin, t: int = 0):
    async def _delayed_task(service: Bannin, t: int):
        await asyncio.sleep(t)
        _logger.info(f"Timer fired for {service.get_name()} @ {datetime.now()}")
        await service.run()

    _t = asyncio.create_task(_delayed_task(service, t))
    timer_tasks[service.get_name()] = {
        "task": _t,
        "cachelife": service.get_config().cachelife,
        "tstamp": datetime.now(tz=UTC),
        "service": service,
        }
    _t.add_done_callback(_clear_task)


def reschedule(service: Bannin, *, for_when: int | None = None):
    _logger.info(f"Rescheduling {service.get_name()}")
    deschedule(service)
    _post_task(
        service,
        t = (
            for_when
            if for_when is not None
            else service.get_config().cachelife
            )
        )

def deschedule(service: Bannin):
    if _td := timer_tasks.pop(service.get_name(), None):
        _td["task"].cancel()




"""Exposes a SearchGit service to find git repos below your home."""

import subprocess
from dataclasses import dataclass, field
from datetime import UTC, datetime
from pathlib import Path

import logistro

from tetsuya._globals import service_types

from . import _base

_logger = logistro.getLogger(__name__)

@dataclass(slots=True)
class Config(_base.Settei):
    cachelife: int = 12 * 60 * 60
    autorefresh: bool = True
    ignore_folders: list[str] = field(default_factory=lambda: [".cache"])
    ignore_paths: list[str] = field(default_factory=list)

@dataclass()
class Report(_base.Tsuho):
    """Report format for SearchGit."""

    retval: int
    stderr: str
    repos: list[Path]

    def long(self) -> str:
        """One full path per line."""
        return "\n".join(str(p.resolve()) for p in sorted(self.repos))

    def short(self) -> str:
        """Comma-separated last directory names."""
        return ", ".join(p.name for p in sorted(self.repos))


class SearchGit(_base.Bannin[Config]):  # is Bannin
    """SearchGit is a class to find git repos below your home directory."""

    @classmethod
    def get_name(cls):
        return cls.__name__

    @classmethod
    def get_config_type(cls):
        return Config

    def __init__(self):
        """Construct a SearchGit service."""

    def _execute(self) -> Report:
        """Execute search of your home repository for git repos."""
        home = Path.home()


        _cfg = self.get_config()
        _logger.info(f"Ignoring folders: {_cfg.ignore_folders}")
        _logger.info(f"Ignoring paths: {_cfg.ignore_paths}")

        # Build the prune expression:
        # ( -path <abs> -o -path <abs> -o -name <nm> -o ... )
        expr = []
        for p in _cfg.ignore_paths:
            expr += ["-path", str(p)]
            expr += ["-o"]
        for name in _cfg.ignore_folders:
            expr += ["-name", str(name)]
            expr += ["-o"]
        # drop last -o, close group, then -prune -o
        expr = [r"(", *expr[:-1], r")", "-prune", "-o"] if expr else []

        cmd = [
            "find",
            str(home),
            *expr,
            "-type",
            "d",
            "-name",
            ".git",
            "-printf",
            r"%h\n",  # print the parent dir of .git
            "-prune",  # and don't descend into the .git dir itself
        ]
        _logger.info(" ".join(cmd))
        p = subprocess.run(  # noqa: S603
            cmd,
            check=False,
            capture_output=True,
        )
        retval, stdout, stderr = p.returncode, p.stdout, p.stderr
        _logger.info("SearchGit ran command `find`.")
        _repos = sorted(
            {Path(p) for p in stdout.decode(errors="ignore").split("\n") if p},
        )
        # we are not skipping things properly
        rep = Report(retval=retval, stderr=stderr.decode(), repos=_repos)
        return rep


service_types.append(SearchGit)



"""Server implements logic for starting and verifying server."""

from __future__ import annotations

import asyncio
import os
import sys
from http import HTTPStatus
from typing import TYPE_CHECKING

import httpx
import logistro
import uvicorn

from tetsuya._globals import active_services, app, cli, service_types
from tetsuya.timer import reconfig

from .client import get_client
from .utils import uds_path

if TYPE_CHECKING:
    from pathlib import Path


_logger = logistro.getLogger(__name__)


def is_server_alive(uds_path: Path) -> bool:
    """Check if server is running."""
    if not uds_path.exists():
        return False
    client: None | httpx.Client = None
    try:
        client = get_client(uds_path, defer_close=False)
        r = client.get("/ping")
        if r.status_code == HTTPStatus.OK:
            _logger.info("Socket ping returned OK- server alive.")
            return True
        else:
            _logger.info(
                f"Socket ping returned {r.status_code}, removing socket.",
            )
            uds_path.unlink()
            return False
    except httpx.TransportError:
        _logger.info("Transport error in socket, removing socket.")
        uds_path.unlink()
        return False
    finally:
        if client:
            client.close()


@cli.command(name="server")
def start():
    """Start the server."""
    async def _start():
        active_services.update(
            {f.__name__: f() for f in service_types},
        )
        if not is_server_alive(p := uds_path()):
            for _n, _s in active_services.items():
                _logger.info(f"Found: {_n}")
                reconfig(_s)
            os.umask(0o077)
            _logger.info("Starting server.")
            server = uvicorn.Server(uvicorn.Config(
                app,
                uds=str(p),          # ✅ same as uvicorn.run(app, uds=str(p))
                loop="asyncio",      # use the current asyncio loop
                lifespan="on",
                reload=False,
                ))
            await server.serve()
        else:
            print("Server already running.", file=sys.stderr)  # noqa: T201
            sys.exit(1)
    asyncio.run(_start())


@app.get("/ping")
def ping():
    """Ping!"""  #  noqa: D400
    _logger.info("Pong!")
    return "pong"
